<?php
// tòmò web analytics
// Copyright since 2011 by Yves Goergen
// Website: http://unclassified.de/go/tomo
// Licence: GNU General Public License (GPL) Version 3
// Encoding: UTF-8 (without BOM) (auto-detect: °°°°°)
//
// This file is derived from:
// UNB Components
// Copyright since 2007 by Yves Goergen
// Website: http://unclassified.de/go/unb2
//
// This file contains the UnbDatabase class.

if (!class_exists('UnbConfiguration', false)) require_once dirname(__FILE__) . '/configuration.class.php';

// The database connection class provides basic query abstraction support and methods to easily
// execute SQL statements and query record sets, records or single values with an SQL query.
//
// It establishes the connection to the database provider with PHP Data Objects (PDO) which
// technically allows using diverse database backends. Still the submitted queries must be
// formulated such that each DBMS to be used understands them.
//
// UnbDatabase is a singleton class, there can only be one single instance of it during runtime.
// This way, the database connection can be used from multiple classes without passing its
// reference everywhere.
//
class UnbDatabase
{
	// Display queries and error messages when an error occurs.
	private $displayQueryErrors = false;
	// Display information for all issued queries. (Ineffective if displayQueryErrors is {{false}}.)
	private $displayQueryInfo = false;

	// Singleton instance
	private static $instance = null;

	public $traceSql = false;

	// Array with information on how to connect to the database
	private $connectData = null;
	// PDO connection
	private $connection = null;
	// PDO driver name for the current database connection
	private $driverName = '';
	// Table prefix, allows storing several UNB tables instances in a single database (schema)
	private $tablePrefix = '';
	// Current transaction nesting level (only one transaction at a time is used in the database backend)
	private $transactionLevel = 0;
	// Remembers whether the database is locked for the current transaction (SQLite only)
	private $databaseLocked = false;
	// Statistics data: Number of queries run in this session
	private $queryCount = 0;
	// Statistics data: Total time that all queries took
	private $totalTime = 0.0;
	// Cached value: Number of rows affected by the last SQL statement (only for DELETE, INSERT and UPDATE)
	private $rowsAffected = 0;
	// Indicates whether the last query caused a constraint violation
	private $constraintError = false;
	// Indicates whether the last query ran against a lock and should be repeated (SQLite only)
	private $busyError = false;
	// Last error message
	private $errorMessage = null;
	// Connection ID (for debugging concurrent connections, not supported by every database driver)
	private $connectionId = 0;
	// Last generated auto value (from auto increment column or implicitly used sequence, triggered
	// by a UnbDbAutoValue parameter)
	private $lastAutoValue = null;

	// Private constructor, does nothing but preventing an unwanted instantiation of this class.
	//
	private function __construct()
	{
	}

	// Gets the instance of this class.
	//
	// If called for the first time, the instance is created.
	//
	// Returns the only UnbDatabase instance
	//
	public static function GetInstance()
	{
		if (!isset(self::$instance))
		{
			self::$instance = new self();
		}
		return self::$instance;
	}

	// Gets a separate instance of this class.
	//
	// This should be used only when needed, like for the logging functions.
	//
	// Returns a new UnbDatabase instance
	//
	public static function GetNewInstance()
	{
		return new self();
	}

	// Enables or disables displaying queries and error messages when an error occurs.
	//
	public function SetDisplayQueryErrors($v)
	{
		$this->displayQueryErrors = $v;
	}

	// Gets a value indicating whether the database connection is open.
	//
	// Returns (bool)
	//
	public function IsOpen()
	{
		return isset($this->connection);
	}

	// Gets the driver name used for the database connection.
	//
	// Returns (string)
	//
	public function GetDriverName()
	{
		return $this->driverName;
	}

	// Gets statistic data.
	//
	// Returns an array with the keys 'queryCount' (number of queries issued) and
	// 'totalTime' (time for all queries in seconds)
	//
	public function GetStatistics()
	{
		return array(
			'queryCount' => $this->queryCount,
			'totalTime' => $this->totalTime);
	}

	// Opens a database connection.
	//
	// connectData = (array) Database connection configuration array containing the following items:
	//                       * 'driver' = PDO driver to be used ('mysql', 'pgsql' or 'sqlite')
	//                       * 'dsn' = DSN string (only with ODBC)
	//                       * 'hostname' = Database server hostname (not with SQLite)
	//                       * 'port' = Database server port number (optional; not with SQLite)
	//                       * 'unix-socket' = Database server UNIX socket (optional; only with MySQL, not with hostname/port)
	//                       * 'dbname' = Database name (not with SQLite)
	//                       * 'username' = Connection login username (not with SQLite)
	//                       * 'password' = Connection login password (not with SQLite)
	//                       * 'filename' = Database filename, relative to this file (only with SQLite)
	//                       * 'schema' = Database schema name (only with pgsql)
	//                       * 'synchronous' = Disk synchronisation mode (optional; 'OFF', 'NORMAL' or 'FULL'; only with SQLite)
	//                       * 'table-prefix' = Table name prefix (optional)
	//
	// Throws InvalidArgumentException
	// Throws UnbDatabaseException
	// Throws PDOException
	//
	public function Open($connectData = null)
	{
		if (isset($connectData))
		{
			if (!is_array($connectData))
				throw new InvalidArgumentException('Invalid argument type: $connectData');
		}
		else
		{
			// No database connection configuration is set, go read it from the global configuration.
			$entries = UnbConfiguration::Get('database.*');
			$connectData = array();
			foreach ($entries as $name => $data)
			{
				$connectData[substr($name, 9)] = $data;
			}
		}

		$this->connectData = $connectData;
		if ($connectData['table-prefix'] != null)
			$this->tablePrefix = $connectData['table-prefix'];

		switch ($this->connectData['driver'])
		{
			case 'mysql':
				$dsn = 'mysql:dbname=' . $this->connectData['dbname'] .
					(isset($this->connectData['hostname']) ? ';host=' . $this->connectData['hostname'] : '') .
					(isset($this->connectData['port']) ? ';port=' . $this->connectData['port'] : '') .
					(isset($this->connectData['unix-socket']) ? ';unix_socket=' . $this->connectData['unix-socket'] : '');
				break;

			case 'odbc':
				$dsn = 'odbc:' . $this->connectData['dsn'];
				break;

			case 'pgsql':
				$dsn = 'pgsql:host=' . $this->connectData['hostname'] .
					(isset($this->connectData['port']) ? ';port=' . $this->connectData['port'] : '') .
					';dbname=' . $this->connectData['dbname'];
				break;

			case 'sqlite':
				if (!file_exists(dirname(__FILE__) . '/' . $this->connectData['filename']))
					throw new UnbDatabaseException('[UNB#37] Database file does not exist');
				if (!filesize(dirname(__FILE__) . '/' . $this->connectData['filename']))
					throw new UnbDatabaseException('[UNB#37] Database file is empty (0 bytes long)');
				$dsn = 'sqlite:' . dirname(__FILE__) . '/' . $this->connectData['filename'];
				break;

			default:
				throw new InvalidArgumentException('Unsupported database driver name. Check your configuration.');
		}
		$this->driverName = $this->connectData['driver'];

		$parameters = array();
		if ($connectData['persistent'])
			$parameters[PDO::ATTR_PERSISTENT] = true;

		$this->connection = new PDO($dsn, $this->connectData['username'], $this->connectData['password'], $parameters);

		// Driver-specific configuration
		switch ($this->driverName)
		{
			case 'mysql':
				$minVersion = '5.0.15';
				$version = $this->QueryValue('SELECT VERSION()');
				if (version_compare($version, $minVersion) < 0)
					throw new Exception('Unsupported MySQL server version ' . $version . '; version ' . $minVersion . ' or newer is required.');

				$this->connection->exec('SET sql_mode = \'ANSI\'');
				$this->connection->exec('SET storage_engine = \'InnoDB\'');
				$this->connection->exec('SET CHARACTER SET \'utf8\'');
				$this->connection->exec('SET NAMES \'utf8\'');
				$this->connectionId = $this->QueryValue('SELECT CONNECTION_ID()');
				break;

			case 'pgsql':
				if (isset($this->connectData['schema']))
					$this->connection->exec('SET search_path TO ' . $this->connection->quote($this->connectData['schema']));
				break;

			case 'sqlite':
				$minVersion = '3.6.19';
				$version = $this->QueryValue('SELECT sqlite_version()');
				if (version_compare($version, $minVersion) < 0)
					throw new Exception('Unsupported SQLite library version ' . $version . '; version ' . $minVersion . ' or newer is required.');

				if (isset($this->connectData['synchronous']))
					$this->connection->exec('PRAGMA synchronous = ' . $this->connectData['synchronous']);
				$this->connection->exec('PRAGMA foreign_keys = ON');
				// Register a user-defined function with SQLite to perform Unicode-safe LOWER()
				// operations on strings. Since PHP comes with such a feature but SQLite does not,
				// we make SQLite use PHP's function here.
				$this->connection->sqliteCreateFunction('LOWER', 'mb_strtolower', 1);
				$this->connection->sqliteCreateFunction('EXTRACTVALUE', array(__CLASS__, 'SQLite_ExtractValue'), 2);
				$this->connection->sqliteCreateFunction('LEVEL', array(__CLASS__, 'SQLite_Level'), 2);
				$this->connection->sqliteCreateFunction('BROWSERNAME', array(__CLASS__, 'SQLite_BrowserName'), 1);
				$this->connection->sqliteCreateFunction('BROWSERVERSION', array(__CLASS__, 'SQLite_BrowserVersion'), 1);
				$this->connection->sqliteCreateFunction('OSNAME', array(__CLASS__, 'SQLite_OsName'), 1);
				$this->connection->sqliteCreateFunction('OSVERSION', array(__CLASS__, 'SQLite_OsVersion'), 1);
				break;
		}

		// Reset statistics after initialisation works
		$this->queryCount = 0;
		$this->totalTime = 0;
	}

	// Closes the database connection.
	//
	public function Close()
	{
		$this->connection = null;
	}

	// SQLite UDF (user defined function) for XML access.
	//
	public static function SQLite_ExtractValue($xml, $xpath)
	{
		$xml = new SimpleXMLElement($xml);
		$data = $xml->xpath($xpath);
		if (is_array($data) && count($data) > 0)
			return $data[0];
		return '';
	}

	// SQLite UDF (user defined function) for number levelling.
	//
	public static function SQLite_Level($num, $levelsStr)
	{
		if (!isset($num)) return null;
		$levels = array_map('intval', explode(',', $levelsStr));
		$li = count($levels) - 1;   // Last index
		if ($num < $levels[0])
		{
			return $levels[0];
		}
		if ($num >= $levels[$li])
		{
			return $levels[$li];
		}
		for ($i = 0; $i < $li; $i++)
		{
			if ($num >= $levels[$i] && $num < $levels[$i + 1])
			{
				$mid = ($levels[$i] + $levels[$i + 1]) / 2;
				if ($num < $mid)
				{
					return $levels[$i];
				}
				else
				{
					return $levels[$i + 1];
				}
			}
		}
		return null;   // Error
	}

	// SQLite UDF (user defined function) for browser grouping.
	//
	public static function SQLite_BrowserName($browser)
	{
		if (!isset($browser)) return null;
		if (preg_match('_^webkit/chrome_', $browser)) return 'Chrome';
		else if (preg_match('_^galeon_', $browser)) return 'Galeon';
		else if (preg_match('_^gecko/(ff|fennec)_', $browser)) return 'Firefox';
		else if (preg_match('_^gecko/nn_', $browser)) return 'Netscape';
		else if (preg_match('_^gecko/seamonkey_', $browser)) return 'SeaMonkey';
		else if (preg_match('_^gecko_', $browser)) return 'Gecko-based';
		else if (preg_match('_^ie/_', $browser)) return 'Internet Explorer';
		else if (preg_match('_^opera/_', $browser)) return 'Opera';
		else if (preg_match('_^webkit/safari_', $browser)) return 'Safari';
		else if (preg_match('_^webkit/omniweb_', $browser)) return 'Omniweb';
		else if (preg_match('_^webkit_', $browser)) return 'Webkit-based';
		else if (preg_match('_^kindle_', $browser)) return 'Kindle';
		else if (preg_match('_^text_', $browser)) return 'Text browser';
		return null;
	}

	// SQLite UDF (user defined function) for browser grouping.
	//
	public static function SQLite_BrowserVersion($browser)
	{
		if (!isset($browser)) return null;
		if (preg_match('_^webkit/chrome/(?:chromium/|frame/)?(\d+)_', $browser, $m)) return $m[1];
		else if (preg_match('_^gecko/ff/(3\.\d+|[4-9]|1\d)_', $browser, $m)) return $m[1];
		else if (preg_match('_^ie/ie/(\d+)_', $browser, $m)) return $m[1];
		else if (preg_match('_^opera/opera/(\d+(\.\d)?)_', $browser, $m)) return $m[1];
		else if (preg_match('_^webkit/safari/(\d+)_', $browser, $m)) return $m[1];
		return null;
	}

	// SQLite UDF (user defined function) for operating system grouping.
	//
	public static function SQLite_OsName($os)
	{
		if (!isset($os)) return null;
		if (preg_match('_^win/phone_', $os)) return 'Windows Phone';
		else if (preg_match('_^win/mobile_', $os)) return 'Windows Mobile';
		else if (preg_match('_^win/ce_', $os)) return 'Windows CE';
		else if (preg_match('_^win_', $os)) return 'Windows';
		else if (preg_match('_^unix/linux/android_', $os)) return 'Android';
		else if (preg_match('_^unix/linux_', $os)) return 'Linux';
		else if (preg_match('_^unix/freebsd_', $os)) return 'FreeBSD';
		else if (preg_match('_^mac/i(pad|phone)_', $os)) return 'iOS';
		else if (preg_match('_^mac/x_', $os)) return 'OSX';
		else if (preg_match('_^blackberry_', $os)) return 'Blackberry';
		else if (preg_match('_^symbian_', $os)) return 'Symbian';
		else if (preg_match('_^unix/sun_', $os)) return 'Sun OS';
		else if (preg_match('_^os2_', $os)) return 'OS/2';
		else if (preg_match('_^java/me_', $os)) return 'Java ME';
		else if (preg_match('_^webos_', $os)) return 'webOS';
		return null;
	}

	// SQLite UDF (user defined function) for operating system grouping.
	//
	public static function SQLite_OsVersion($os)
	{
		if (!isset($os)) return null;
		if (preg_match('_^win/(phone|mobile)/([0-9.]+)_', $os, $m)) return $m[2];
		else if (preg_match('_^win/(95|98|me)_', $os, $m)) return strtoupper($m[1]);
		else if (preg_match('_^win/nt/4_', $os, $m)) return 'NT 4';
		else if (preg_match('_^win/nt/(2000|xp|2003|7|8)_', $os, $m)) return strtoupper($m[1]);
		else if (preg_match('_^win/nt/vista_', $os, $m)) return 'Vista';
		else if (preg_match('_^unix/linux/android/([0-9.]+)_', $os, $m)) return $m[1];
		else if (preg_match('_^mac/i(pad|phone)/([0-9.]+)_', $os, $m)) return $m[2];
		else if (preg_match('_^mac/x/([0-9.]+)_', $os, $m)) return $m[1];
		return null;
	}

	// Shows query information and errors.
	//
	// ps = (PDO/PDOStatement) Prepared statement or PDO connection used to retrieve the error details
	// r = (bool) Result value. {{true}}: The query was successful. {{false}}: There was an error that should be displayed.
	// parameters = (array) Bound parameters used to execute the SQL query.
	//                      Those will be inserted for every "?" when displaying the SQL query.
	// time = (float) Time in seconds it took to execute the query.
	//
	// Uses UnbDatabase::$displayQueryInfo
	//
	private function ShowQueryInfo($ps, $r, $parameters, $time)
	{
		if (!$r || $this->displayQueryInfo)
		{
			echo '<div class="sql-info" onclick="alert(&quot;' . htmlspecialchars(str_replace("\n", '\n', addslashes(UnbDebug::GetCallTrace(1)))) . '&quot;)"><b>SQL execute:</b> ';
			if ($ps instanceof PDOStatement)
			{
				$sql = htmlspecialchars($ps->queryString, ENT_NOQUOTES);
			}
			else if ($ps instanceof PDO)
			{
				$sql = htmlspecialchars($parameters, ENT_NOQUOTES);
				$parameters = null;
			}
			if ($parameters)
			{
				foreach ($parameters as $p)
				{
					if (!isset($p))
						$p = 'NULL';
					else if (is_bool($p))
						$p = $p ? '1' : '0';
					else if (!is_int($p))
						$p = '\'' . htmlspecialchars(mb_substr($p, 0, 30), ENT_NOQUOTES) . '\'';
					$sql = preg_replace('_\?_', '<i>' . $p . '</i>', $sql, 1);
				}
			}
			echo $sql;
			if ($time >= 0)
			{
				echo ' &ndash; ' . ($time >= 0.005 ? '<b>' : '') . round($time * 1000, 1) . 'ms' . ($time >= 0.005 ? '</b>' : '');
			}
			echo '</div>';
		}

		if (!$r)
		{
			// $ps may be a PDOStatement or a PDO object.
			$ei = $ps->errorInfo();
			echo '<div class="sql-error"><b>SQL error:</b> [SQLSTATE:' . $ei[0] . ', ' . $ei[1] . '] ' . htmlspecialchars($ei[2], ENT_NOQUOTES) . '</div>';
		}
	}

	// Finds out what the problem was. :-)
	//
	// ps = (PDO/PDOStatement) Prepared statement or PDO connection used to retrieve the error details
	//
	private function DetermineErrorCondition($ps)
	{
		// Reset error condition
		$this->constraintError = false;
		$this->busyError = false;

		$ei = $ps->errorInfo();

		// Evaluate SQLSTATE value, class "23": Constraint violations
		if (substr($ei[0], 0, 2) == '23')
			$this->constraintError = true;

		// Driver-specific error codes
		switch ($this->driverName)
		{
			case 'sqlite':
				if ($ei[1] == 5)   // SQLITE_BUSY = 5
				{
					$this->busyError = true;
					// Wait for a while so that the query can then be repeated immediately
					usleep(mt_rand(2000, 20000));   // 2 ... 20 ms
				}
				else if ($ei[1] == 19)   // SQLITE_CONSTRAINT = 19
				{
					$this->constraintError = true;
				}
				break;
		}

		$this->errorMessage = $ei[2];
	}

	// Gets a value indicating whether the last error condition describes a constraint violation.
	//
	// This value is only valid if the previous query actually failed.
	//
	// Returns (bool) value
	//
	public function IsConstraintError()
	{
		return $this->constraintError;
	}

	// Gets the message of the last occured database error.
	//
	// Returns (string) Error message
	//
	public function GetErrorMessage()
	{
		return $this->errorMessage;
	}

	// Optimises the database.
	//
	// extensive = (bool) Set this to {{true}}, in order to optimise the database more thoroughly.
	//                    This may take a much longer time for large databases. If {{false}}, merely
	//                    the index statistics are updated which can improve performance but
	//                    does not save disk space.
	//
	// Returns {{true}}, if successful, {{false}} otherwise.
	//
	public function Optimise($extensive = false)
	{
		$r = true;
		switch ($this->driverName)
		{
			case 'mysql':
				if ($extensive)
				{
					$r &= $this->Query('OPTIMIZE TABLE "#action", "#session", "#text", "#visitor"');
					// For InnoDB: OPTIMIZE TABLE rebuilds the table to update index statistics
					// and free unused space in the clustered index
				}
				else
				{
					$r &= $this->Query('ANALYZE TABLE "#action", "#session", "#text", "#visitor"');
					// ANALYZE TABLE analyzes and stores the key distribution for a table.
					// During the analysis, the table is locked with a write lock for InnoDB.
				}
				break;
			case 'pgsql':
				if ($extensive)
				{
					$r &= $this->Execute('VACUUM FULL');
					$r &= $this->Execute('REINDEX DATABASE ?', $this->connectData['database']);
				}
				else
				{
					$r &= $this->Execute('VACUUM');
					$r &= $this->Execute('ANALYZE');
				}
				break;
			case 'sqlite':
				if ($extensive)
				{
					$r &= $this->Execute('VACUUM');
					// VACUUM rebuilds the database file. This can be delayed (blocking) when a
					// transaction is active.
					$r &= $this->Execute('ANALYZE');
				}
				else
				{
					$r &= $this->Execute('ANALYZE');
				}
				break;
		}
		return $r;
	}

	// Starts a new transaction.
	//
	// If a transaction is already started, only the counter is increased by 1, but no real nested
	// transaction is started.
	//
	// lock = (bool) Lock the entire database. This option is only regarded for database system
	//               that don't support table or row-level locking, like SQLite. It can only be
	//               used when no other transaction is currently started (i.e. the transaction
	//               level counter is 0).
	//
	// Throws BadMethodCallException when trying to acquire a lock for the transaction if a
	// transaction has already started.
	//
	// Returns {{true}}, if successful, {{false}} otherwise.
	//
	public function StartTransaction($lock = false)
	{
		if (!isset($this->connection))
			throw new BadMethodCallException('Database not opened');
		if ($lock && $this->transactionLevel > 0 && !$this->databaseLocked)
			throw new BadMethodCallException('You cannot lock during a transaction that has no lock.');

		$r = true;
		if ($this->transactionLevel == 0)
		{
			if ($this->traceSql) UnbDebug::Trace('  SQL: BEGIN TRANSACTION');
			$start = microtime(true);
			$retries = 200;   // avg. 2 seconds for busy errors
			do
			{
				if ($lock)
				{
					// Driver-specific configuration
					switch ($this->driverName)
					{
						case 'sqlite':
							ignore_user_abort(true);
							$ps = $this->connection->prepare('BEGIN EXCLUSIVE TRANSACTION');
							$r &= $ps->execute();
							break;
						default:
							$r &= $this->connection->beginTransaction();
							break;
					}
					$this->databaseLocked = true;
				}
				else
				{
					$r &= $this->connection->beginTransaction();
				}
				if (!$r) $this->DetermineErrorCondition($this->connection);
				$retries--;
			}
			while (!$r && $this->busyError && $retries > 0);   // SQLite busy loop workaround
			if (!$r && $this->busyError)
				throw new UnbDatabaseException('Database is busy too long. Failed starting transaction.');
			$time = microtime(true) - $start;
			$this->totalTime += $time;
			if ($this->displayQueryErrors)
				$this->ShowQueryInfo($this->connection, $r, 'START TRANSACTION', $time);
		}
		else if ($this->displayQueryInfo)
		{
			echo '<div class="sql-info" onclick="alert(&quot;' . htmlspecialchars(str_replace("\n", '\n', addslashes(UnbDebug::GetCallTrace()))) . '&quot;)"><b>UnbDatabase operation:</b> Called StartTransaction, changed transaction level from ' . $this->transactionLevel . ' to ' . ($this->transactionLevel + 1) . '.</div>';
		}
		$this->transactionLevel++;
		return $r;
	}

	// Rolls back a transaction.
	//
	// If the transaction counter is greater than 1, the transaction is not actually rolled back.
	// It is the callers' responsibility to pass on the error condition and call RollBack() as
	// often as required to really roll back the transaction.
	//
	// Throws BadMethodCallException if not in a transaction.
	//
	// Returns {{true}}, if successful, {{false}} otherwise.
	//
	public function RollBack()
	{
		if (!isset($this->connection))
			throw new BadMethodCallException('Database not opened');
		if ($this->transactionLevel == 0)
			throw new BadMethodCallException('No transaction is active');

		$r = true;
		$this->transactionLevel--;
		if ($this->transactionLevel == 0)
		{
			if ($this->traceSql) UnbDebug::Trace('  SQL: ROLLBACK');
			$start = microtime(true);
			if ($this->databaseLocked)
			{
				// Driver-specific configuration
				switch ($this->driverName)
				{
					case 'sqlite':
						// Transactions not started through the PDO API cannot be ended through it.
						$ps = $this->connection->prepare('ROLLBACK');
						$r &= $ps->execute();
						ignore_user_abort(false);
						break;
					default:
						$r &= $this->connection->rollBack();
						break;
				}
			}
			else
			{
				$r &= $this->connection->rollBack();
			}
			$time = microtime(true) - $start;
			$this->totalTime += $time;
			if ($this->displayQueryErrors)
				$this->ShowQueryInfo($this->connection, $r, 'ROLLBACK', $time);
			$this->databaseLocked = false;
		}
		else if ($this->displayQueryInfo)
		{
			echo '<div class="sql-info" onclick="alert(&quot;' . htmlspecialchars(str_replace("\n", '\n', addslashes(UnbDebug::GetCallTrace()))) . '&quot;)"><b>UnbDatabase operation:</b> Called Rollback, changed transaction level from ' . ($this->transactionLevel + 1) . ' to ' . $this->transactionLevel . '.</div>';
		}
		return $r;
	}

	// Commits a transaction.
	//
	// If the transaction counter is greater than 1, the transaction is not actually commited.
	//
	// Throws BadMethodCallException if not in a transaction.
	//
	// Returns {{true}}, if successful, {{false}} otherwise.
	//
	public function Commit()
	{
		if (!isset($this->connection))
			throw new BadMethodCallException('Database not opened');
		if ($this->transactionLevel == 0)
			throw new BadMethodCallException('No transaction is active');

		$r = true;
		$this->transactionLevel--;
		if ($this->transactionLevel == 0)
		{
			if ($this->traceSql) UnbDebug::Trace('  SQL: COMMIT');
			$start = microtime(true);
			if ($this->databaseLocked)
			{
				// Driver-specific configuration
				switch ($this->driverName)
				{
					case 'sqlite':
						// Transactions not started through the PDO API cannot be ended through it.
						$ps = $this->connection->prepare('COMMIT');
						$r &= $ps->execute();
						if ($r) ignore_user_abort(false);   // Committing can fail. Transactions are then rolled back.
						break;
					default:
						$r &= $this->connection->commit();
						break;
				}
			}
			else
			{
				$r &= $this->connection->commit();
			}
			$time = microtime(true) - $start;
			$this->totalTime += $time;
			if (!$r) $this->DetermineErrorCondition($this->connection);
			if ($this->displayQueryErrors)
				$this->ShowQueryInfo($this->connection, $r, 'COMMIT', $time);
			$this->databaseLocked = false;
		}
		else if ($this->displayQueryInfo)
		{
			echo '<div class="sql-info" onclick="alert(&quot;' . htmlspecialchars(str_replace("\n", '\n', addslashes(UnbDebug::GetCallTrace()))) . '&quot;)"><b>UnbDatabase operation:</b> Called Commit, changed transaction level from ' . ($this->transactionLevel + 1) . ' to ' . $this->transactionLevel . '.</div>';
		}
		return $r;
	}

	// Gets the next value of a sequence.
	//
	// table = (string) Table name for which the sequence was created.
	//
	// Not all database systems support sequences. Those who don't will use auto increment values
	// instead. This function returns null for them.
	//
	private function GetSequenceValue($table)
	{
		// Driver-specific configuration
		switch ($this->driverName)
		{
			case 'pgsql':
				return $this->QueryValue('nextval(\'' . $this->tablePrefix . $table . '_seq\')');
		}
		return null;
	}

	// Resets the last auto value.
	//
	private function ResetLastAutoValue()
	{
		$this->lastAutoValue = null;
	}

	// Fetches the last inserted auto increment value from the database connection and stores it
	// for later retrieval with GetLastAutoValue.
	//
	// Not all database systems support auto values. Those who don't will have to explicitly get a
	// sequence value and assign it before executing the query. This function does nothing for
	// them.
	//
	private function FetchLastAutoValue()
	{
		// Driver-specific configuration
		switch ($this->driverName)
		{
			case 'mysql':
			case 'sqlite':
				$this->lastAutoValue = intval($this->connection->lastInsertId());
				break;
			// NOTE: MSSQL needs a special handling, see http://de.php.net/manual/de/pdo.lastinsertid.php#87589
		}
	}

	// Gets the auto value generated by the last query.
	//
	// This will only be set if the last query has set a UnbDbAutoValue parameter.
	//
	// Returns (int) Last generated auto-increment/sequence value.
	//
	public function GetLastAutoValue()
	{
		return $this->lastAutoValue;
	}

	// Basic SQL query processing.
	//
	// This method strips out SQL comments (they could crash PHP/Apache during development) and
	// replaces the "#" placeholder with the specified table name prefix (if any).
	//
	// sql = (string) SQL query to process
	//
	// Returns the processed and (if appropriate) modified SQL query.
	//
	private function ProcessSql($sql)
	{
		// Driver-specific configuration
		switch ($this->driverName)
		{
			case 'sqlite':
				// SELECT ... FOR UPDATE is not supported. Instead, the database should be locked
				// for the transaction.
				if (preg_match('_^\s*SELECT\s+.+(\s+FOR\s+UPDATE)$_is', $sql, $m))
				{
					if (!$this->databaseLocked && $this->displayQueryErrors)
					{
						echo '<div style="color: gray; font-size: 0.85em;"><b>SQL execute:</b> ' . htmlspecialchars($sql) . '</div>';
						echo '<div style="color: red; font-size: 0.85em;"><b>SQL warning:</b> A SELECT ... FOR UPDATE query was issued but the database was not locked for the transaction.</div>';
					}
					$sql = substr($sql, 0, -strlen($m[1]));
				}
				break;
		}

		return preg_replace('_\s*--\s.*?$_m', '',
			preg_replace('_(?<=")#(?=[a-zA-Z])_', $this->tablePrefix, $sql));
	}

	// Prepares an SQL statement.
	//
	// sql = (string) SQL statement.
	//
	// Throws InvalidArgumentException on invalid argument type.
	// Throws BadMethodCallException if the database is not opened.
	// Throws UnbDatabaseException if the statement could not be prepared.
	//
	// Returns a PDOStatement object
	//
	public function Prepare($sql)
	{
		if (!is_string($sql))
			throw new InvalidArgumentException('Invalid argument type: $sql');
		if (!isset($this->connection))
			throw new BadMethodCallException('Database not opened');

		$sql = $this->ProcessSql($sql);
		$ps = $this->connection->prepare($sql);
		if (!$ps)
		{
			$this->DetermineErrorCondition($this->connection);
			$this->ShowQueryInfo($this->connection, false, $sql, 0);
			throw new UnbDatabaseException('Statement could not be prepared. ' . $this->errorMessage);
		}
		return $ps;
	}

	// Actually executes an SQL query.
	//
	// sql = (string) SQL statement
	//       (PDOStatement) Already prepared statement
	// parameters = (array) Parameters to be used in this SQL statement, {{null}} if no parameters are used.
	//                      This can only be (an array of) scalar values, UNB's objects are not allowed!
	//
	// Throws InvalidArgumentException on invalid argument type.
	// Throws BadMethodCallException if the database is not opened.
	// Throws UnbDatabaseException if the statement could not be prepared or executed.
	//
	// Returns the PDOStatement object to be used for further data retrieval.
	//
	private function InternalExecute($sql, $parameters)
	{
		if (!isset($this->connection))
			throw new BadMethodCallException('Database not opened');

		$start = microtime(true);
		// Put single parameter argument into an array
		if (!is_array($parameters) && isset($parameters))
		{
			$parameters = array($parameters);
		}
		// Prepare SQL statement if not yet done
		if (is_string($sql))
		{
			// DEBUG: Resolve all parameter bindings, until that works well (PHP PDO bug, no bug# yet)
			if (count($parameters))
			{
				$pos = 0;
				while (($pos = strpos($sql, '?', $pos)) !== false)
				{
					if (!count($parameters))
					{
						throw new InvalidArgumentException('Too few parameters specified for query.');
					}
					$tmp = substr($sql, 0, $pos);
					$param = array_shift($parameters);
					if (!isset($param))
					{
						$tmp .= 'NULL';
					}
					else if ($param instanceof UnbDbAutoValue)   // NOTE: SQLite only!
					{
						$hasAutoValue = true;
						$tmp .= 'NULL';
					}
					else
					{
						$tmp .= $this->Quote($param);
					}
					$sql = $tmp . substr($sql, $pos + 1);
					$pos = strlen($tmp);
				}
				if (count($parameters))
				{
					throw new InvalidArgumentException('Too many parameters specified for query (' . count($parameters) . ' too many).');
				}
			}

			$ps = $this->Prepare($sql);
		}
		else if ($sql instanceof PDOStatement)
		{
			$ps = $sql;
			$sql = $ps->queryString;
		}
		else
		{
			throw new InvalidArgumentException('Invalid argument type: $sql');
		}
		// Handle auto values
		$this->ResetLastAutoValue();
#		$hasAutoValue = false;
#		if (is_array($parameters))
#		{
#			foreach ($parameters as &$param)
#			{
#				if ($param instanceof UnbDbAutoValue)
#				{
#					if (!$hasAutoValue)
#					{
#						$hasAutoValue = true;
#						$param = $this->GetSequenceValue($param->GetTable());
#					}
#					else
#					{
#						throw new InvalidArgumentException('More than one auto value passed as parameter.');
#					}
#				}
#			}
#		}
		// Execute the statement
		if ($this->traceSql) UnbDebug::Trace('  SQL: ' . str_replace("\n", "\n" . str_repeat(' ', 44), $sql));
		$retries = 200;   // avg. 2 seconds for busy errors
		do
		{
			for ($i = 0; $i < count($parameters); $i++)
			{
				$v = $parameters[$i];
				if (!isset($v)) $type = PDO::PARAM_NULL;
				else if (is_int($v)) $type = PDO::PARAM_INT;
				#else if (is_float($v)) $type = PDO::PARAM_FLOAT;
				else $type = PDO::PARAM_STR;
				#echo 'x' . ($i + 1) . 'x' . $v . 'xxxxxxxx' . $type . '<br />';
				$ps->bindParam($i + 1, $v, $type);
			}
			$r = $ps->execute();
			if (!$r) $this->DetermineErrorCondition($ps);
			$retries--;
		}
		while (!$r && $this->busyError && $retries > 0);   // SQLite busy loop workaround
		if (!$r && $this->busyError)
			throw new UnbDatabaseException('Database is busy too long. Failed statement: ' . $sql);
		$time = microtime(true) - $start;
		$this->totalTime += $time;
		$this->queryCount++;
		if ($this->displayQueryErrors)
			$this->ShowQueryInfo($ps, $r, $parameters, $time);
		if (!$r)
			throw new UnbDatabaseException('Statement could not be executed. ' . $this->errorMessage);
		// Get last auto value
		if ($hasAutoValue)
			$this->FetchLastAutoValue();
		return $ps;
	}

	// Executes an SQL statement that returns no data.
	//
	// sql = (string) SQL statement
	//       (PDOStatement) Already prepared statement
	// parameters = (array) Parameters to be used in this SQL statement, {{null}} if no parameters are used.
	//
	// Throws UnbDatabaseException if the statement could not be prepared or executed.
	//
	public function Execute($sql, $parameters = null)
	{
		$ps = $this->InternalExecute($sql, $parameters);
		if (isset($ps))
			$this->rowsAffected = $ps->rowCount();
		else
			$this->rowsAffected = -1;
	}

	// Inserts a single record into a table.
	//
	// table = (string) Table to insert the record in (without prefix)
	// values = (array) Array with the column names as keys and record data as values.
	//
	public function Insert($table, $values)
	{
		$sqlCols = '';
		$sqlValues = '';
		foreach ($values as $key => $value)
		{
			$sqlCols .= ($sqlCols ? ', ' : '') . '"' . $key . '"';
			$sqlValues .= ($sqlValues ? ', ' : '') . '?';
		}
		$this->Execute(
			'INSERT INTO "#' . $table . '" (' . $sqlCols . ') VALUES (' . $sqlValues . ')',
			array_values($values));
	}

	// Updates a record in a table.
	//
	// table = (string) Table to update the record in (without prefix)
	// values = (array) Array with the column names as keys and record data as values.
	// condSql = (string) Record condition to check, will be used verbatim in the WHERE part.
	// condParams = (any/array) Parameters for each "?" in the condition.
	//
	public function Update($table, $values, $condSql, $condParams = null)
	{
		$sql = '';
		foreach ($values as $key => $value)
		{
			$sql .= ($sql ? ', ' : '') . '"' . $key . '" = ?';
		}
		$values = array_values($values);
		if (is_array($condParams) && count($condParams))
		{
			$values = array_merge($values, $condParams);
		}
		else if (isset($condParams))
		{
			$values = array_merge($values, array($condParams));
		}
		$this->Execute(
			'UPDATE "#' . $table . '" SET ' . $sql . ' WHERE ' . $condSql,
			$values);
	}

	// Deletes a record from a table.
	//
	// table = (string) Table to delete the record from (without prefix)
	// condSql = (string) Record condition to check, will be used verbatim in the WHERE part.
	// condParams = (any/array) Parameters for each "?" in the condition.
	//
	public function Delete($table, $condSql, $condParams = null)
	{
		$this->Execute(
			'DELETE FROM "#' . $table . '" WHERE ' . $condSql,
			$condParams);
	}

	// Executes an SQL statement and returns all records.
	//
	// sql = (string) SQL statement
	//       (PDOStatement) Already prepared statement
	// parameters = (array) Parameters to be used in this SQL statement, {{null}} if no parameters are used.
	// fetchStyle = PDO fetch style, see PHP manual (mostly PDO::FETCH_NUM or PDO::FETCH_ASSOC).
	//
	// Throws UnbDatabaseException if the statement could not be prepared or executed.
	//
	// Returns an array of all records
	//
	public function Query($sql, $parameters = null, $fetchStyle = PDO::FETCH_ASSOC)
	{
		$ps = $this->InternalExecute($sql, $parameters);
		$records = $ps->fetchAll($fetchStyle);
		return $records;
	}

	// Executes an SQL statement and returns the first column of all records.
	//
	// sql = (string) SQL statement
	//       (PDOStatement) Already prepared statement
	// parameters = (array) Parameters to be used in this SQL statement, {{null}} if no parameters are used.
	//
	// Throws UnbDatabaseException if the statement could not be prepared or executed.
	//
	// Returns an array of all records' first column
	//
	public function QueryArray($sql, $parameters = null)
	{
		$ps = $this->InternalExecute($sql, $parameters);
		$records = $ps->fetchAll(PDO::FETCH_NUM);
		$array = array();
		foreach ($records as $record)
		{
			$array[] = $record[0];
		}
		return $array;
	}

	// Executes an SQL statement and returns the first record.
	//
	// sql = (string) SQL statement
	//       (PDOStatement) Already prepared statement
	// parameters = (array) Parameters to be used in this SQL statement, {{null}} if no parameters are used.
	// fetchStyle = PDO fetch style, see PHP manual (mostly PDO::FETCH_NUM or PDO::FETCH_ASSOC).
	//
	// Throws UnbDatabaseException if the statement could not be prepared or executed.
	//
	// Returns the first record
	//
	public function QueryRecord($sql, $parameters = null, $fetchStyle = PDO::FETCH_ASSOC)
	{
		$ps = $this->InternalExecute($sql, $parameters);
		$record = $ps->fetch($fetchStyle);
		$ps->closeCursor();
		return $record;
	}

	// Executes an SQL statement and returns a single value.
	//
	// sql = (string) SQL statement
	//       (PDOStatement) Already prepared statement
	// parameters = (array) Parameters to be used in this SQL statement, {{null}} if no parameters are used.
	//
	// Throws UnbDatabaseException if the statement could not be prepared or executed.
	//
	// Returns the first record's first column, if successful, {{null}} if no record was fetched.
	//
	public function QueryValue($sql, $parameters = null)
	{
		$ps = $this->InternalExecute($sql, $parameters);
		$value = $ps->fetchColumn();
		$ps->closeCursor();
		if ($value === false) return null;
		return $value;
	}

	// Gets the number of rows affected by the last SQL statement (only for DELETE, INSERT and UPDATE).
	//
	// Returns (int) number of rows
	//
	public function GetRowsAffected()
	{
		return $this->rowsAffected;
	}

	// Converts a date/time value between PHP's UNIX timestamps and the SQL notation.
	//
	// in = (int) UNIX timestamp
	//      (string) SQL notation in the format "YYYY-MM-DD HH:MM:SS" and UT timezone.
	//               Only a date like "YYYY-MM-DD" is also accepted.
	//
	// Throws InvalidArgumentException on unrecognised date/time format or invalid argument type.
	//
	// Returns the date/time in the respective other format. An input value of (int) 0
	// or (string) "0000-00-00 00:00:00" is both interpreted as "no value" and converted
	// accordingly.
	//
	public static function ConvertDateTime($in)
	{
		if (!isset($in))
		{
			return null;
		}
		if (is_int($in))
		{
			if ($in == 0) return '0000-00-00 00:00:00';
			return gmdate('Y-m-d H:i:s', $in);
		}
		if (is_string($in))
		{
			if (preg_match('_^(\d{4})-(\d{2})-(\d{2})(?: (\d{2}):(\d{2}):(\d{2})(\.\d+)?)?$_', $in, $m))
			{
				if (intval($m[1]) == 0) return 0;   // No content (year is 0)
				return gmmktime(intval($m[4]), intval($m[5]), intval($m[6]), intval($m[2]), intval($m[3]), intval($m[1]));
			}
			throw new InvalidArgumentException('Unrecognised date/time format');
		}
		throw new InvalidArgumentException('Invalid argument type: $in');
	}

	// Escape a value to use the wildcards * and ? instead of SQL's % and _.
	//
	// This value must be used in a LIKE ? ESCAPE ? context, with \ as the escape value.
	//
	// $s = (string) The value to escape
	//
	// Returns (string) the escaped value
	//
	public static function EscapeLikeWildcards($s)
	{
		return str_replace(
			array('%',   '_',   '\\',   '*', '?'),
			array('\\%', '\\_', '\\\\', '%', '_'),
			$s);
	}

	// Interprets a mixed value returned from the database as boolean.
	//
	// s = (any) Raw value
	//
	// Returns (bool) A boolean interpretation of that value.
	//
	public static function AsBool($s)
	{
		$s = strtolower(trim($s));
		return $s !== '' &&
			$s !== '0' &&
			$s !== 'false' &&
			$s !== 'no' &&
			$s !== 'off';
	}

	public function Quote($s)
	{
		if (is_int($s) || is_float($s))
		{
			return $s;
		}
		else // string
		{
			#return $this->connection->quote($s);   // NOTE: This function gives unprecise results and doesn't respect the data type
			return "'" . str_replace("'", "''", $s) . "'";
		}
	}
} // class UnbDatabase

// Represents an auto-increment/sequenca value in a database query.
//
// Use an instance of this class as query parameter whenever the actual value stored in the
// database shall be automatically incremented. Only one query parameter of this type is allowed
// in a single query.
//
class UnbDbAutoValue
{
	// The database table the auto value is requested for.
	private $table;

	// Class constructor.
	//
	// Some DBMS that use native sequences require a sequence name to get the next value of. This
	// name is derived from the table name that is specified in this constructor.
	//
	// table = (string) Name of the database table the auto value is requested for.
	//
	public function __construct($table)
	{
		$this->table = $table;
	}

	// Gets the table name the auto value is requested for.
	//
	// Returns (string) Table name
	//
	public function GetTable()
	{
		return $this->table;
	}
} // UnbDbAutoValue

// The database exception is meant to be used for errors in the database layer, like invalid or
// failed SQL queries or database connection problems.
//
class UnbDatabaseException extends Exception
{
} // class UnbDatabaseException

?>
